import os
import sys
import unittest

from ctypes import *
from ctypes.util import find_library
from ctypes.decorators import cdecl, name_library

# specify which libraries the decorator has to load.  Normally, one
# would use ful so-names to specify exact the library we want.

if os.name == "nt":
    name_library("libc", "msvcrt")
    name_library("libm", "msvcrt")
elif os.name == "darwin":
    name_library("libc", find_library("c"))
    name_library("libm", find_library("m"))
else:
    name_library("libc", "libc.so.6")
    name_library("libm", find_library("m"))

# load libraries with cdll magic.  I still don't know how to find
# 'libc.so.6' from 'c' on linux, otherwise this could be a lot nicer.

if os.name == "nt":
    libm = libc = cdll.msvcrt
elif sys.platform == "darwin":
    libc = cdll.c
    libm = cdll.m
else:
    libc = CDLL("libc.so.6")
    libm = cdll.m


class LibTest(unittest.TestCase):
    def test_sqrt(self):
        libm.sqrt.argtypes = c_double,
        libm.sqrt.restype = c_double
        self.failUnlessEqual(libm.sqrt(4.0), 2.0)
        import math
        self.failUnlessEqual(libm.sqrt(2.0), math.sqrt(2.0))

    def test_cdecl_decorator(self):

        # @ cdecl(c_double, "libm", [c_double])
        def sqrt(value):
            return sqrt._api_(value)
        # Oh well, compatibility with Python 2.3
        sqrt = cdecl(c_double, "libm", [c_double]) (sqrt)

        self.failUnlessEqual(sqrt(4.0), 2.0)

    def test_qsort(self):
        comparefunc = CFUNCTYPE(c_int, POINTER(c_char), POINTER(c_char))
        libc.qsort.argtypes = c_void_p, c_int, c_int, comparefunc
        libc.qsort.restype = None

        def sort(a, b):
            return cmp(a[0], b[0])

        chars = create_string_buffer("spam, spam, and spam")
        libc.qsort(chars, len(chars)-1, sizeof(c_char), comparefunc(sort))
        self.failUnlessEqual(chars.raw, "   ,,aaaadmmmnpppsss\x00")

    def test_qsort_decorator(self):
        from ctypes.decorators import cdecl, name_library

        CMPFUNC = CFUNCTYPE(c_int, c_void_p, c_void_p)
        #@ cdecl(None, "libc", [c_void_p, c_int, c_int, CMPFUNC])
                               
        def qsort(sequence, cmp_func):
            itemsize = sizeof(sequence) / len(sequence)
            qsort._api_(sequence, len(sequence), itemsize, CMPFUNC(cmp_func))
        qsort = cdecl(None, "libc", [c_void_p, c_int, c_int, CMPFUNC])(qsort)

        def sortfunc(a, b):
            a = cast(a, POINTER(c_int))
            b = cast(b, POINTER(c_int))
            return cmp(a[0], b[0])

        ints = (c_int * 9)(1, 3, 5, 7, 9, 8, 6, 4, 2)
        qsort(ints, sortfunc)
        self.failUnlessEqual(ints[:], [1, 2, 3, 4, 5, 6, 7, 8, 9])

if __name__ == "__main__":
    unittest.main()
